import sys
import itertools
import types

from ._compat import string_types

PY26 = sys.version_info < (2, 7)

_registration_id = itertools.count()

_token_registrations = {}


class Registration(object):

    def __init__(self, func, hook, token=None, tags=None, needs=None, provides=None):
        super(Registration, self).__init__()
        self.id = next(_registration_id)
        self.hook = hook
        self.func = func
        self.token = token
        self.needs = _normalize_deps(needs)
        self.provides = _normalize_deps(provides)
        self.tags = set(tags) if tags else None
        if not isinstance(func, (classmethod, staticmethod, types.MethodType)) and not hasattr(func, "gossip"):
            func.gossip = self

        self.valid = True

    def invalidate(self):
        self.valid = False

    def has_tags(self, tags):
        if tags is None:
            return True
        if self.tags is None:
            return True
        return bool(set(tags) & self.tags)

    def unregister(self):
        if self.hook is not None:
            self.hook.unregister(self)
            assert self.hook is None

    def is_active(self):
        return self.hook is not None

    def can_be_called(self):
        return True

    def __call__(self, *args, **kwargs):
        assert self.valid
        return self.func(*args, **kwargs)

    def __repr__(self):
        return repr(self.func)


def _normalize_deps(deps):
    if not deps:
        deps = ()
    if isinstance(deps, string_types):
        deps = [deps]
    return frozenset(deps)



from munch import Munch
from .utils import TestSteps, _noop

import pytest
import gossip
from gossip.exceptions import UndefinedHook, UnsupportedHookTags


steps = TestSteps()

@steps.add
def _make_group_strict(ctx):
    raises_context = _noop()
    if ctx.registered_hook:
        if not ctx.defined_hook:
            raises_context = pytest.raises(UndefinedHook)
        elif not ctx.is_tags_valid_for_strict_group:
            raises_context = pytest.raises(UnsupportedHookTags)
    with raises_context:
        gossip.get_or_create_group("group").set_strict()
    ctx.made_strict = True

@steps.add
def _make_define_hook(ctx):
    gossip.define("group.defined_hook", tags=ctx.defined_tags)
    ctx.defined_hook = True

@steps.add
def _make_register_hook(ctx):
    raises_context = _noop()
    if ctx.made_strict:
        if not ctx.defined_hook:
            raises_context = pytest.raises(UndefinedHook)
        elif not ctx.is_tags_valid_for_strict_group:
            raises_context = pytest.raises(UnsupportedHookTags)

    with raises_context:
        @gossip.register("group.defined_hook", tags=ctx.registeration_tags)
        def handler():
            pass
    ctx.registered_hook = True

def _clean_workspace():
    gossip.registry.hooks = {}
    gossip.registry.groups = {None: gossip.registry.groups[None]}

@pytest.mark.parametrize("tags_info", [
    (None, None, True),
    (None, ("fake_tag",), False),
    (None, ("fake_tag", "fake_tag2"), False),
    (tuple(), None, True),
    (tuple(), ("fake_tag",), False),
    (tuple(), ("fake_tag", "fake_tag2"), False),
    (("some_tag",), None, True),
    (("some_tag",), tuple(), True),
    (("some_tag",), ("some_tag",), True),
    (("some_tag",), ("fake_tag",), False),
    (("some_tag",), ("some_tag", "fake_tag"), False),
    (("some_tag",), ("fake_tag", "fake_tag2"), False),
    (("some_tag", "other_tag"), None, True),
    (("some_tag", "other_tag"), tuple(), True),
    (("some_tag", "other_tag"), ("some_tag",), True),
    (("some_tag", "other_tag"), ("fake_tag",), False),
    (("some_tag", "other_tag"), ("some_tag", "fake_tag"), False),
    (("some_tag", "other_tag"), ("some_tag", "other_tag"), True),
    (("some_tag", "other_tag"), ("fake_tag", "fake_tag2"), False),
    ])
@steps.get_all_permutations
def test_tags(tags_info, steps):
    defined_tags, registered_tags, valid_strict_tags = tags_info
    ctx = Munch(
        made_strict=False,
        defined_hook=False,
        registered_hook=False,
        defined_tags=defined_tags,
        registeration_tags=registered_tags,
        is_tags_valid_for_strict_group=valid_strict_tags,
        )
    _clean_workspace()
    for step_func in steps:
        step_func(ctx)

def test_single_set_tags_per_hook():
    some_hook = gossip.define('group.some_hook')
    other_hook = gossip.define('group.other_hook', tags=("other_tag",))
    some_hook.set_tags(('some_tag',))

    with pytest.raises(AssertionError):
        some_hook.set_tags(('some_other_tag',))

    with pytest.raises(AssertionError):
        other_hook.set_tags(('some_other_tag',))

def test_trigger_hook_with_unsupported_tags():
    gossip.define('group.some_hook', tags=("some_tag",))
    gossip.trigger('group.some_hook', tags=("fake_tag",))
    gossip.get_or_create_group('group').set_strict()
    with pytest.raises(UnsupportedHookTags):
        gossip.trigger_with_tags('group.some_hook', tags=("fake_tag",))


@pytest.mark.parametrize("hook_name", ["main_group1.some_hook", "main_group2.sub_group.other_hook"])
def test_set_strict_with_subgroups_and_tags(hook_name):
    def handler():
        pass

    main_group_name = hook_name.partition(".")[0]
    gossip.register(handler, hook_name)

    with pytest.raises(UndefinedHook):
        gossip.get_or_create_group(main_group_name).set_strict()
    hook = gossip.define(hook_name, tags=("some_tag", "other_tag"))

    illegal_tags_hook = gossip.define(".".join([hook.group.full_name, "hook_with_illegal_tags"]))
    gossip.register(handler, illegal_tags_hook.full_name, tags=("fake_tag"))
    with pytest.raises(UnsupportedHookTags):
        gossip.get_or_create_group(main_group_name).set_strict()
    illegal_tags_hook.unregister_all()

    gossip.get_or_create_group(main_group_name).set_strict()
    with pytest.raises(UnsupportedHookTags):
        @gossip.register(hook_name, tags=("fake_tag",))
        def handler2():
            pass
    with pytest.raises(UnsupportedHookTags):
        gossip.trigger_with_tags(hook_name, tags=("fake_tag",))
    gossip.trigger_with_tags(hook_name, tags=("some_tag", "other_tag"))



from . import registry
from ._compat import itervalues, iteritems
from .exception_policy import ExceptionPolicy, Inherit, RaiseImmediately
from .exceptions import GroupNotFound, NameAlreadyUsed
from .helpers import DONT_CARE


class Group(object):

    def __init__(self, name, parent=None):
        super(Group, self).__init__()
        self.name = name
        self._parent = parent
        if self._parent is not None and not self._parent.is_global():
            self.full_name = "{0}.{1}".format(
                self._parent.full_name, self.name)
        else:
            self.full_name = name
        self.reset()

    def is_global(self):
        return self._parent is None

    def is_strict(self):
        return self._strict

    def undefine(self):
        self.undefine_children()
        if self._parent is not None:
            self._parent.remove_child(self.name)
        registry.groups.pop(self.full_name)

    def reset(self):
        if hasattr(self, "_children"):
            self.undefine_children()
        self._strict = False
        self._unconstrained_handler_priority = DONT_CARE
        self._children = {}
        self._parent_exception_policy = None
        self.set_exception_policy(
            RaiseImmediately() if self._parent is None else Inherit())

    def set_unconstrained_handler_priority(self, priority):
        """Controls when handlers without needs/provides specifications should be fired
        relative to handlers with needs/provides.

        :param priority: ``gossip.FIRST`` means that unconstrained handlers should be fired first, ``gossip.LAST`` means last
        """
        self._unconstrained_handler_priority = priority
        for hook in self.iter_hooks():
            hook.recompute_call_order()
        for group in self.iter_subgroups():
            group.set_unconstrained_handler_priority(priority)

    def get_unconstrained_handler_priority(self):
        return self._unconstrained_handler_priority

    def undefine_children(self):
        for child in list(itervalues(self._children)):
            child.undefine()

    def set_strict(self, strict=True):
        """Marks this group as a strict group, meaning all hooks registered must be defined in advance

        :param strict: controls whether or not this group should be turned to strict
        """
        for child in itervalues(self._children):
            if isinstance(child, Group):
                child.set_strict(strict)
            elif strict:
                child.validate_strict()
        self._strict = strict

    def get_undefined_hooks(self):
        returned = [
            hook for hook in self.get_hooks() if not hook.is_defined()]
        for group in self.get_subgroups():
            returned.extend(group.get_undefined_hooks())
        return returned

    def notify_parent_exception_policy_changed(self):
        self._parent_exception_policy = None

    def add_child(self, name, child):
        self._children[name] = child
        return child

    def remove_child(self, child):
        self._children.pop(child)

    def get_or_create_subgroup(self, name):
        returned = self._subgroups.get(name)
        if returned is None:
            returned = self._subgroups[name] = Group(name, parent=self)
        return returned

    def get_subgroups(self):
        return list(self.iter_subgroups())

    def iter_subgroups(self):
        for child in itervalues(self._children):
            if isinstance(child, Group):
                yield child

    def get_hooks(self):
        return list(self.iter_hooks())

    def iter_hooks(self):
        for child in itervalues(self._children):
            if isinstance(child, Group):
                continue
            yield child

    def remove_all_children(self):
        self._children.clear()

    def unregister_token(self, token):
        """Unregisters all handlers that were registered with ``token`` in this group
        """
        # todo: optimize this
        for hook in self.iter_hooks():
            for registration in hook.get_registrations():
                if registration.token == token:
                    registration.invalidate()
                    registration.unregister()
        for group in self.iter_subgroups():
            group.unregister_token(token)

    def set_exception_policy(self, policy):
        """ Determines how exceptions are handled by hooks in this group
        """
        if not isinstance(policy, ExceptionPolicy):
            raise ValueError(
                "Expected ExceptionPolicy instance. Got {0!r}".format(policy))
        if isinstance(policy, Inherit) and self._parent is None:
            raise RuntimeError(
                "Global gossip group cannot inherit exception policy from parent")

        self._exception_policy = policy
        for subgroup in self.iter_subgroups():
            subgroup.notify_parent_exception_policy_changed()

    def get_exception_policy(self):
        """Returns the current exception policy to be used by this group
        """
        if isinstance(self._exception_policy, Inherit):
            returned = self._parent_exception_policy
            if returned is None:
                returned = self._parent_exception_policy = self._parent.get_exception_policy(
                )
            return returned
        return self._exception_policy

    def __repr__(self):
        return "<Gossip group {0!r}>".format(self.name)


def get_global_group():
    """Gets the global (or root) group
    """
    return get_group(None)


def get_groups():
    """Gets all defined groups (including the global group)
    """
    return list(group for group_name, group in iteritems(registry.groups) if group_name is not None)


def get_or_create_group(name):
    """Tries to retrieve an existing group, and if it doesn't exist, create a new group
    """
    try:
        return get_group(name)
    except GroupNotFound:
        return create_group(name)


def get_group(name):
    """Returns an existing group with a given name

    :raises: KeyError if no such group exists
    """
    try:
        return registry.groups[name]
    except KeyError:
        raise GroupNotFound(name)


def create_group(name):
    """Creates and returns a new named group

    :rtype: :class:`gossip.group.Group`
    """
    if name in registry.groups:
        raise NameAlreadyUsed(
            "Group with name {0} already exists".format(name))

    if name in registry.hooks is not None:
        raise NameAlreadyUsed(
            "Hook with name {0} already exists. Cannot create group with same name".format(name))

    if "." in name:
        parent_name, group_name = name.rsplit(".", 1)
        parent = get_or_create_group(parent_name)
    else:
        parent = get_global_group()
        group_name = name

    group = Group(group_name, parent=parent)
    registry.groups[name] = group
    parent.add_child(group_name, group)
    return group


def unregister_token(token):
    """Shortcut for :func:`gossip.groups.Group.unregister_token` on the global group
    """
    get_global_group().unregister_token(token)

# Create the global group
registry.groups[None] = Group("**GLOBAL**")



import itertools
from uuid import uuid4

import logbook
import gossip
import gossip.hooks
import pytest

hook_id = itertools.count()


@pytest.fixture(autouse=True, scope='session')
def setup_logging():
    logbook.StderrHandler().push_application()
    logbook.Flags(errors='raise').push_application()

@pytest.fixture(autouse=True, scope="function")
def clear_registrations():
    for hook in gossip.get_all_hooks():
        hook.unregister_all()

@pytest.fixture(autouse=True, scope="function")
def undefine_all():
    gossip.get_global_group().reset()

@pytest.fixture
def hook_name():
    return "hook{0}".format(next(hook_id))

@pytest.fixture
def hook(hook_name):
    return gossip.define(hook_name)

@pytest.fixture
def registered_hook(hook_name):
    return RegisteredHook(hook_name)

@pytest.fixture
def registered_hooks():
    returned = []
    for i in range(10):
        if i % 3 == 0:
            hook_name = "group{0}.hook{1}".format(next(hook_id), next(hook_id))
        else:
            hook_name = "hook{0}".format(next(hook_id))
        returned.append(RegisteredHook(hook_name))
    return returned

timestamp = itertools.count()

class RegisteredHook(object):

    def __init__(self, hook_name):
        super(RegisteredHook, self).__init__()

        self.name = hook_name
        self._fail = False
        self.kwargs = {"a": 1, "b": 2, "c": 3}
        self.num_called = 0
        self._dependency = None

        class HandlerException(Exception):
            pass

        self.exception_class = HandlerException

        def handler(**kw):
            assert kw == self.kwargs
            gossip.wait_for(self._dependency is None or self._dependency.called)
            if self._dependency is not None:
                assert self._dependency.called
            self.num_called += 1
            self.last_timestamp = next(timestamp)
            if self._fail:
                raise self.exception_class()

        self.func = handler

        gossip.register(func=handler, hook_name=hook_name)

    def depend_on(self, hook):
        self._dependency = hook

    @property
    def called(self):
        return self.num_called > 0

    def fail_when_called(self):
        self._fail = True

    def works(self):
        old_num_caled = self.num_called
        self.trigger()
        return self.num_called == old_num_caled + 1

    def trigger(self):
        gossip.trigger(self.name, **self.kwargs)


@pytest.fixture
def token():
    return str(uuid4())


@pytest.fixture
def checkpoint():
    return Checkpoint()


class Checkpoint(object):

    num_times = 0

    def __call__(self):
        self.num_times += 1

    @property
    def called(self):
        return self.num_times > 0

    def reset(self):
        self.num_times = 0


class Timeline(object):

    def __init__(self):
        super(Timeline, self).__init__()
        self.hook_name = 'parent_group_{0}.subgroup_{0}.hook_{1}'.format(uuid4(), uuid4())
        self.timestamps = itertools.count(1000)
        self.event_index = itertools.count()

    def get_group(self):
        return gossip.get_group(self.hook_name.rsplit('.', 1)[0])

    def get_parent_group(self):
        return gossip.get_group(self.hook_name.split('.')[0])

    def register(self, **kwargs):
        evt = Event(next(self.event_index))
        @gossip.register(self.hook_name, **kwargs)
        def callback():
            evt.timestamp = next(self.timestamps)
        return evt

    def trigger(self):
        gossip.trigger(self.hook_name)

@pytest.fixture
def timeline():
    return Timeline()

class Event(object):

    _timestamp = None

    def __init__(self, index):
        super(Event, self).__init__()
        self.index = index

    def __repr__(self):
        return '<Evt {0}>'.format(self.index)

    @property
    def timestamp(self):
        assert self._timestamp is not None
        return self._timestamp

    @timestamp.setter
    def timestamp(self, t):
        assert self._timestamp is None
        self._timestamp = t



import logbook

import gossip


def test_deprecated_hook():
    with logbook.TestHandler(force_heavy_init=True) as h:

        hook = gossip.define('hook', deprecated=True)
        assert h.records == []

        @hook.register
        def handler():
            pass

        [r] = h.records

        assert r.filename == __file__



import pytest
from gossip.exceptions import CannotResolveDependencies
from gossip.utils import _topological_sort


def test_topological_sort(indices, graph, expected):
    sorted = _topological_sort(indices, set(graph))
    assert len(sorted) == len(indices)
    assert set(sorted) == set(indices)
    assert sorted == expected


def test_cycles():
    with pytest.raises(CannotResolveDependencies):
        _topological_sort([0, 1, 2, 3], set([(0, 1), (1, 2), (2, 0)]))


@pytest.fixture(params=[
    (set([(0, 1), (1, 2)]), [0, 1, 2, 3, 4]),
    (set([(4, 0), (2, 4), (3, 2)]), [1, 3, 2, 4, 0]),
    (set([(4, 0), (2, 4), (3, 1), (3, 2)]), [3, 2, 4, 0, 1]),
    ])
def graph_expected(request):
    return request.param

@pytest.fixture
def graph(graph_expected):
    return graph_expected[0]

@pytest.fixture
def expected(graph_expected):
    return graph_expected[1]


@pytest.fixture
def indices():
    return list(range(5))



import logbook
from contextlib import contextmanager

from ._compat import reraise

_logger = logbook.Logger(__name__)

class ExceptionPolicy(object):

    @contextmanager
    def context(self):
        ctx = TriggerContext()
        try:
            yield ctx
        except:
            _logger.debug('Caught exception while calling hooks', exc_info=True)
            raise
        self._handle_trigger_end(ctx)

    def handle_exception(self, ctx, exc_info):
        ctx.exceptions.append(exc_info)

    def _handle_trigger_end(self, ctx):
        pass

class RaiseImmediately(ExceptionPolicy):

    def handle_exception(self, ctx, exc_info):
        reraise(exc_info[0], exc_info[1], exc_info[2])

class RaiseDefer(ExceptionPolicy):

    def _handle_trigger_end(self, ctx):
        if ctx.exceptions:
            exc_type, exc_value, exc_tb = ctx.exceptions[0]
            reraise(exc_type, exc_value, exc_tb)

class IgnoreExceptions(ExceptionPolicy):
    pass

class Inherit(ExceptionPolicy):
    pass

class TriggerContext(object):

    def __init__(self):
        super(TriggerContext, self).__init__()
        self.exceptions = []



import gossip
import pytest

_HOOK_NAME = 'some_group.some_hook'
_ACCEPTED_TAGS = ('tag1', 'tag2')
_INCOMPATIBLE_TAG_SETS = (
    ('tag11',),
    ('tag1', 'tag2', 'unknown'),
)


def test_trigger_hook(specify_tags):
    if specify_tags:
        gossip.trigger_with_tags(_HOOK_NAME, tags=_ACCEPTED_TAGS)
    else:
        gossip.trigger_with_tags(_HOOK_NAME, tags=())


@pytest.mark.parametrize('tags', _INCOMPATIBLE_TAG_SETS)
def test_trigger_with_bad_tags(tags, is_strict):
    if not is_strict:
        return
    with pytest.raises(gossip.exceptions.UnsupportedHookTags):
        gossip.trigger_with_tags(_HOOK_NAME, tags=tags)


@pytest.mark.parametrize('tags', _INCOMPATIBLE_TAG_SETS)
def test_register_with_bad_tags(tags, is_strict):
    if not is_strict:
        return

    def func():
        pass

    with pytest.raises(gossip.exceptions.UnsupportedHookTags):
        func = gossip.register(_HOOK_NAME, tags=tags)(func)

@pytest.fixture(params=[True, False])
def register_before_define(request):
    return request.param

@pytest.fixture(params=[True, False])
def create_group_before_define(request):
    return request.param

@pytest.fixture(params=[True, False, 'strict'])
def is_defined(request):
    return request.param

@pytest.fixture
def is_strict(is_defined):
    returned = not isinstance(is_defined, bool)
    assert not returned or is_defined == 'strict'
    return returned

@pytest.fixture(params=[True, False])
def specify_tags(request):
    return request.param


@pytest.fixture(autouse=True)
def setup_hook(is_defined, is_strict, specify_tags, register_before_define, create_group_before_define):
    # note: the combination of register before define *and* create strict group
    # before define is guaranteed to create an error
    if register_before_define:
        if not is_strict or not create_group_before_define:
            gossip.register(_HOOK_NAME)(lambda: None)

    if create_group_before_define:
        _get_or_create_group(is_strict)

    if is_defined:
        if specify_tags:
            gossip.define(_HOOK_NAME, tags=_ACCEPTED_TAGS)
        else:
            gossip.define(_HOOK_NAME)

    if not create_group_before_define:
        _get_or_create_group(is_strict)

def _get_or_create_group(is_strict):
    g = gossip.get_or_create_group(_HOOK_NAME.split('.')[0])
    if is_strict:
        g.set_strict()



class HookException(Exception):
    pass

class NameAlreadyUsed(HookException):
    pass

class HookNotFound(LookupError):
    pass

class GroupNotFound(LookupError):
    pass

class NotNowException(Exception):
    pass

class CannotResolveDependencies(Exception):
    pass

class UndefinedHook(Exception):
    pass

class UnsupportedHookTags(Exception):
    pass



import functools
import logbook
import sys
from contextlib import contextmanager
from types import GeneratorType

from . import groups, registry
from ._compat import itervalues, string_types
from .exceptions import (CannotResolveDependencies, HookNotFound,
                         NameAlreadyUsed, NotNowException, UndefinedHook,
                         UnsupportedHookTags)
from .registration import Registration
from .utils import topological_sort_registrations

from logbook.utils import log_deprecation_message

_logger = logbook.Logger(__name__)


class Hook(object):

    def __init__(self, group, name, arg_names=(), doc=None, deprecated=False):
        super(Hook, self).__init__()
        self.group = group
        self.name = name
        self.tags = None
        if self.group.is_global():
            self.full_name = name
        else:
            self.full_name = "{0}.{1}".format(self.group.full_name, self.name)
        registry.hooks[self.full_name] = self
        self._registrations = []
        self._arg_names = arg_names
        self._trigger_internal_hooks = self.full_name != "gossip.on_handler_exception"
        self._defined = False
        self._pre_trigger_callbacks = []
        self._unmet_deps = frozenset()
        self.doc = doc
        self.deprecated = deprecated

    def add_pre_trigger_callback(self, callback):
        self._pre_trigger_callbacks.append(callback)
        return callback

    def remove_pre_trigger_callback(self, callback):
        self._pre_trigger_callbacks.remove(callback)

    def undefine(self):
        self.group.remove_child(self.name)
        registry.hooks.pop(self.full_name)

    def set_tags(self, tags):
        assert not self.tags, "Cannot override exists tags {0} with {1}".format(
            self.tags, tags)
        self.tags = tags

    def get_registrations(self):
        return list(self._registrations)

    def mark_defined(self):
        self._defined = True

    def is_defined(self):
        return self._defined

    def get_argument_names(self):
        return self._arg_names

    def __call__(self, **kwargs):
        return self.trigger(kwargs)

    def validate_tags(self, tags, is_strict=None):
        if is_strict is None:
            is_strict = self.group.is_strict()
        if not is_strict:
            return
        normalize_tags = lambda tag_: set(tag_) if tag_ is not None else set()
        extra_tags = normalize_tags(tags) - normalize_tags(self.tags)
        if extra_tags:
            raise UnsupportedHookTags("hook {0} support {1} tags, not: {2}".format(
                self.full_name, self.tags, extra_tags))

    def validate_strict(self, registrations_to_validate=None):
        if not self._defined:
            raise UndefinedHook(
                "hook {0} wasn't defined yet".format(self.full_name))
        if registrations_to_validate is None:
            registrations_to_validate = self._registrations
        for registration in registrations_to_validate:
            self.validate_tags(registration.tags, is_strict=True)

    def register(self, func, token=None, tags=None, needs=None, provides=None):
        """Registers a new handler to this hook
        """
        if self.deprecated:
            log_deprecation_message('Hook {0} is deprecated!'.format(self.full_name), frame_correction=+1)
        returned = Registration(func, self, token=token, tags=tags, needs=needs, provides=provides)
        if self.group.is_strict():
            self.validate_strict([returned])
        self._registrations.append(returned)
        if returned.needs or returned.provides:
            try:
                self.recompute_call_order()
            except:
                self._registrations.pop()
                raise
        return returned

    def recompute_call_order(self):
        self._registrations = topological_sort_registrations(self._registrations, unconstrained_priority=self.group.get_unconstrained_handler_priority())
        self._unmet_deps = frozenset(n for r in self._registrations for n in r.needs) - frozenset(p for r in self._registrations for p in r.provides)


    def unregister(self, registration):
        assert registration.hook is self
        self._registrations.remove(registration)
        registration.hook = None

    def unregister_all(self):
        del self._registrations[:]

    def trigger(self, kwargs, tags=None):
        if self._unmet_deps:
            raise CannotResolveDependencies('Hook {0!r} has unmet dependencies: {1}'.format(self, ', '.join(map(str, self._unmet_deps))))
        if self.full_name in _muted_stack[-1]:
            _logger.debug("Hook {0!r} muted, skipping trigger", self)
            return

        self.validate_tags(tags)
        exception_policy = self.group.get_exception_policy()

        registrations = self._registrations
        deferred = []

        with exception_policy.context() as ctx:
            while True:
                any_resolved = False
                for registration in list(registrations):
                    if not registration.valid:
                        continue
                    if not registration.has_tags(tags):
                        continue
                    try:
                        exc_info = self._call_registration(
                            registration, kwargs)
                    except NotNowException:
                        deferred.append(registration)
                        continue
                    else:
                        any_resolved = True
                    if exc_info is not None:
                        exception_policy.handle_exception(ctx, exc_info)
                if deferred:
                    if not any_resolved:
                        raise CannotResolveDependencies(
                            "Cannot resolve handler dependencies for {0}".format(self))
                    registrations = deferred
                    deferred = []
                else:
                    break

    def _call_registration(self, registration, kwargs):
        exc_info = None
        for callback in self._pre_trigger_callbacks:
            callback(registration, kwargs)
        try:
            registration(**kwargs)  # pylint: disable=star-args
        except NotNowException:
            raise
        except:
            exc_info = sys.exc_info()
            if self._trigger_internal_hooks:
                trigger("gossip.on_handler_exception",
                        handler=registration.func, exception=exc_info, hook=self)
            _logger.debug("Exception occurred while calling %s",
                         registration, exc_info=exc_info)
            # TODO: debug here
        return exc_info

    def __repr__(self):
        return "<Hook {0}({1})>".format(self.name, ", ".join(self._arg_names))


def trigger(hook_name, **kwargs):
    """Triggers a hook by name, causing all of its handlers to be called
    """
    trigger_with_tags(hook_name, kwargs, None)


def trigger_with_tags(hook_name, kwargs=None, tags=None):
    """Triggers a hook by name, specifying the tags to use when triggering.
    If the hook receives keyword arguments, they have to be passed using the *kwargs* parameter, and not via
    direct keyword arguments
    """
    hook = registry.hooks.get(hook_name)
    if hook is not None:
        hook.trigger(kwargs or {}, tags)


def get_or_create_hook(hook_name, **kwargs):
    try:
        return get_hook(hook_name)
    except HookNotFound:
        return create_hook(hook_name, **kwargs)


def get_hook(hook_name):
    """Gets a hook by its name

    :raises: :class:`gossip.exceptions.HookNotFound` if the hook wasn't defined already
    """
    try:
        return registry.hooks[hook_name]
    except KeyError:
        raise HookNotFound("Hook {0} does not exist".format(hook_name))


def create_hook(hook_name, **kwargs):
    """Creates a hook with the given full name, creating intermediate groups if necessary
    """
    if hook_name in registry.hooks:
        raise NameAlreadyUsed(
            "A hook named {0} already exists. Cannot create a hook with the same name".format(hook_name))

    if hook_name in registry.groups:
        raise NameAlreadyUsed(
            "A group named {0} already exists. Cannot create a hook with the same name".format(hook_name))

    if "." in hook_name:
        group_name, hook_base_name = hook_name.rsplit(".", 1)
    else:
        group_name = None
        hook_base_name = hook_name

    group = groups.get_or_create_group(group_name)
    hook = Hook(group, hook_base_name, **kwargs)
    group.add_child(hook_base_name, hook)
    return hook


def define(hook_name, **kwargs):
    """Defines a new hook with the given name

    :returns: The :class:`gossip.hook.Hook` object created
    """
    tags = kwargs.pop('tags', None)
    returned = get_or_create_hook(hook_name, **kwargs)
    if returned.is_defined():
        raise NameAlreadyUsed("Hook {0} is already defined".format(hook_name))
    if tags:
        returned.set_tags(tags)
    returned.mark_defined()
    return returned


def register(func=None, hook_name=None, token=None, tags=None, needs=None, provides=None):
    """Registers a new function to a hook

    :param hook_name: full name of hook to register to
    :param token: token to register with. This can be used to later unregister a group of handlers which have a
           specific token
    :params needs: list of keywords (strings) that this registration needs, causing any hook that provides any of them to happen before this registration
    :params provides: list of keywords (strings) that this registration provides
    :returns: The function (for decorator chaining)

    """
    if isinstance(func, string_types):
        assert hook_name is None
        hook_name = func
        func = None

    if func is None:
        return functools.partial(register, hook_name=hook_name, token=token, tags=tags, needs=needs, provides=provides)
    assert hook_name is not None
    registration = get_or_create_hook(
        hook_name).register(func, token=token, tags=tags, needs=needs, provides=provides)
    assert registration
    return func


def unregister_all(hook_name):
    """
    Unregisters all handlers from the given hook name
    """
    try:
        hook = get_hook(hook_name)
    except HookNotFound:
        return
    hook.unregister_all()


def get_all_hooks():
    return list(itervalues(registry.hooks))

def get_all_registrations():
    return [reg for hook in get_all_hooks()
            for reg in hook.get_registrations()]


_muted_stack = [set()]


@contextmanager
def mute_context(hook_names):
    """A context manager, during the execution of which the specified hook names will be ignored. Any code that tries to
    trigger these hooks will trigger no callback.

    :type hook_names: list, tuple or generator of full hook names to mute
    """
    if not isinstance(hook_names, (list, tuple, GeneratorType)):
        raise TypeError('hook names to mute must be a list or a tuple')
    _muted_stack.append(_muted_stack[-1] | set(hook_names))
    try:
        yield
    finally:
        _muted_stack.pop(-1)



from .exception_policy import IgnoreExceptions, RaiseDefer, RaiseImmediately
from .exceptions import NotNowException
from .groups import (create_group, get_global_group, get_group, get_groups,
                     get_or_create_group, unregister_token)
from .hooks import (define, get_all_hooks, get_all_registrations, get_hook, Hook, register, trigger,
                    trigger_with_tags, mute_context
                    )
from .blueprint import Blueprint
from .helpers import FIRST, DONT_CARE, LAST


def set_exception_policy(policy):
    """This is a shortcut to calling :func:`gossip.group.Group.set_exception_policy` on the global
    hook group"""
    return get_global_group().set_exception_policy(policy)

def not_now():
    """Defers execution of the current handler to later, after other handlers have been called.

    This is useful for depending on other hook handlers
    """
    raise NotNowException()

def wait_for(cond):
    """Defers execution of the current handler (see :func:`not_now`) if ``cond`` is False-y.
    """
    if not cond:
        not_now()



from uuid import uuid4

from . import hooks
from . import groups


class Blueprint(object):
    """Represents a hook blueprint

    Blueprints are sets of hooks that can be registered or unregistered with a single operation.
    """

    def __init__(self):
        super(Blueprint, self).__init__()
        self._hooks = []
        self._token = str(uuid4())

    def register(self, hook_name, **kwargs):
        """Registers a function to this blueprint

        .. seealso:: :func:`gossip.register`
        """
        def decorator(func):
            self._hooks.append((hook_name, func, kwargs))
            return func
        return decorator

    def install(self, group=None):
        """Installs the blueprint, registering all of its handlers to their respective hooks
        """
        try:
            for hook_name, func, kwargs in self._hooks:
                if group is not None:
                    hook_name = '{0}.{1}'.format(group, hook_name)
                hooks.register(func=func, hook_name=hook_name, token=self._token, **kwargs)
        except:
            self.uninstall()
            raise

    def uninstall(self):
        """Uninstalls the blueprint, unregistering all of its hooks
        """
        groups.unregister_token(self._token)



import pytest

import gossip
from gossip.exceptions import CannotResolveDependencies

def test_needs_provides_simple(timeline):
    evt1 = timeline.register(needs=['something'])
    evt2 = timeline.register(provides=['something'])
    timeline.trigger()
    assert evt1.timestamp > evt2.timestamp


def test_needs_provides_complex(timeline):
    events = [
        timeline.register(provides=['0'], needs=['1']),
        timeline.register(provides=['1'], needs=['2', '3']),
        timeline.register(provides=['2'], needs=['5']),
        timeline.register(provides=['3'], needs=['5']),
        timeline.register(provides=['4'], needs=['0']),
        timeline.register(provides=['5']),
    ]
    timeline.trigger()

    assert sorted(events, key=lambda e: e.timestamp) == [
        events[5], events[3], events[2], events[1], events[0], events[4]
    ]


def test_needs_provides_cyclic_simple(timeline):
    timeline.register(needs=['1'], provides=['0'])
    with pytest.raises(CannotResolveDependencies):
        timeline.register(needs=['0'], provides=['1'])
    # make sure the cyclic registration was not added
    assert len(gossip.get_hook(timeline.hook_name)._registrations) == 1


def test_needs_provides_cyclic_complex(timeline):
    timeline.register(needs=['1'], provides=['0'])
    timeline.register(needs=['2'], provides=['1'])
    timeline.register(needs=['3'], provides=['2'])
    timeline.register(needs=['4'], provides=['3'])
    with pytest.raises(CannotResolveDependencies):
        timeline.register(needs=['0'], provides=['4'])
    # make sure the cyclic registration was not added
    assert len(gossip.get_hook(timeline.hook_name)._registrations) == 4


@pytest.mark.parametrize('priority', [gossip.FIRST, gossip.DONT_CARE, gossip.LAST])
@pytest.mark.parametrize('use_parent_group', [True, False])
def test_needs_provides_unconstrained_priority(timeline, priority, use_parent_group):
    dontcare = timeline.register()
    evt1 = timeline.register(needs=['a'], provides=['b'])
    evt2 = timeline.register(provides=['a'])
    if use_parent_group:
        group = timeline.get_parent_group()
    else:
        group = timeline.get_group()

    group.set_unconstrained_handler_priority(priority)
    timeline.trigger()

    if priority in (gossip.FIRST, gossip.DONT_CARE):
        assert dontcare.timestamp < evt2.timestamp < evt1.timestamp
    else:
        assert evt2.timestamp < evt1.timestamp < dontcare.timestamp


def test_policy_gets_reset():
    gossip.get_global_group().set_unconstrained_handler_priority(gossip.FIRST)
    gossip.get_global_group().reset()
    assert gossip.get_global_group().get_unconstrained_handler_priority() == gossip.DONT_CARE


def test_unmet_dependencies(timeline):
    evt = timeline.register(needs=['a'])
    with pytest.raises(CannotResolveDependencies):
        timeline.trigger()

    evt2 = timeline.register(provides=['a'])
    timeline.trigger() # ok
